package com.sillycycle.bagleyd.cubes;

/*
 * @(#)CubesFrame.java
 *
 * Copyright 1994 - 2022  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Label;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Clip;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.UnsupportedAudioFileException;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JTextField;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.sillycycle.bagleyd.util.ArgumentParser;
import com.sillycycle.bagleyd.util.ColumnLayout;
import com.sillycycle.bagleyd.util.ComponentUtil;
import com.sillycycle.bagleyd.util.Icon;
import com.sillycycle.bagleyd.util.IODialog;
import com.sillycycle.bagleyd.util.OKDialog;
import com.sillycycle.bagleyd.util.Spinner;

/**
 * The <code>CubesFrame</code> class is to be run as an application.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

public class CubesFrame extends JFrame implements ActionListener {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");
	static final int MAX_BLOCKS = 8;
	static final int MIN_SPEED = 1;
	static final int MAX_SPEED = 50;
	static final int NEVER = -1;
	static final String SCOREFILE = "cubes.scores";
	static final String[] COMMAND_STRINGS = new String[] {
		"Get Buffer", "Write Buffer", "Show Buffer",
		"Undo", "Redo", "Clear",
		"Randomize", "Auto-solve",
		"Number",
		"Speed >", "Slow <", "Sound @",
		"Description ?", "Features !", "References ^", "About",
		"Get File", "Write File", "Exit"};
	static final char[] COMMAND_CHARS = new char[] {
		'g', 'w', '*',
		'u', 'r', 'c', 'z', 's',
		'n', '>', '<', '@',
		'?', '!', '^', 'a', 'G', 'W', 'x'};

	static ArgumentParser argumentParser = null;
	Icon mediumIcon = null;
	Icon largeIcon = null;
	int widthValue = 516;
	int heightValue = 516;
	static int xValue = 0;
	static int yValue = 0;
	Color blockColor = Color.lightGray;
	Color fgColor = Color.cyan;
	Color bgColor = Color.gray;
	boolean sound = false;
	boolean controlValue = true;
	boolean numberValue = true;
	CubesSize blocksValue = new CubesSize();
	int baseValue = CubesInterface.DEFAULT_BASE;
	int delayValue = 10;
	int speedValue = MAX_SPEED + MIN_SPEED - 1 - delayValue;
	boolean debug = false;
	boolean randomized = false;
	JPanel topPanel = new JPanel();
	Cubes puzzle;
	int movesDisplay = 0;
	String recordDisplay = "NEVER";
	String messageDisplay = "Play Cubes! (use mouse or keypad)";
	JTextField movesValueLabel = new JTextField(Integer.toString(
		movesDisplay));
	JTextField recordValueLabel = new JTextField(recordDisplay);
	Label messageLabel = new Label(messageDisplay, Label.CENTER);
	// JLabel constantly resizes
	OKDialog aDialog = null;
	StringBuilder stringBuilderSave;
	IODialog ioDialog = null;
	int[][][] puzzleRecord = null;
	JMenuBar menuBar;
	JPanel mainPanel = new JPanel();
	JPanel secondaryPanel = new JPanel();
	JFrame frame;
	//JCheckBoxMenuItem numberMenuItem = null;
	JCheckBoxMenuItem soundMenuItem = null;
	//Slider blockXChanger = null, blockYChanger = null, blockZChanger = null;
	//Slider speedChanger = null;
	Spinner blockXChanger = null, blockYChanger = null, blockZChanger = null;
	Spinner speedChanger = null;

	public String writePuzzleString() {
		return this.stringBuilderSave.toString();
	}

	public void getPuzzleString(String s) {
		puzzle.stringSave = s;
	}

	public void processKey(char key, boolean control) {
		puzzle.processKey(key, control);
	}

	public boolean getToggleSound() {
		return sound;
	}

	public void toggleSound() {
		sound = !sound;
		if (controlValue)
			soundMenuItem.setSelected(sound);
	}

	public String getOpt(String string) {
		return argumentParser.getOption(string);
	}

	public int getOpt(String param, int value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			return Integer.parseInt(string);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public boolean getOpt(String param, boolean value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			if (string.equalsIgnoreCase("true") ||
					string.equalsIgnoreCase("yes") ||
					string.equalsIgnoreCase("on"))
				return true;
			else if (string.equalsIgnoreCase("false") ||
					string.equalsIgnoreCase("no") ||
					string.equalsIgnoreCase("off"))
				return false;
			return (Integer.parseInt(string) != 0);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public Color getOpt(String param, Color value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			return Color.decode(string);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public void getOpts() {
		widthValue = getOpt("windowWidth", widthValue);
		heightValue = getOpt("windowHeight", heightValue);
		xValue = getOpt("windowX", xValue);
		yValue = getOpt("windowY", yValue);
		blockColor = getOpt("blockColor", blockColor);
		fgColor = getOpt("fgColor", fgColor);
		bgColor = getOpt("bgColor", bgColor);
		sound = getOpt("sound", sound);
		controlValue = getOpt("control", controlValue);
		blocksValue.x = getOpt("sizex", blocksValue.x);
		if (blocksValue.x < CubesInterface.MIN_BLOCKS)
			blocksValue.x = CubesInterface.DEFAULT_BLOCKSX;
		blocksValue.y = getOpt("sizey", blocksValue.y);
		if (blocksValue.y < CubesInterface.MIN_BLOCKS)
			blocksValue.y = CubesInterface.DEFAULT_BLOCKSY;
		blocksValue.z = getOpt("sizez", blocksValue.z);
		if (blocksValue.z < CubesInterface.MIN_BLOCKS)
			blocksValue.z = CubesInterface.DEFAULT_BLOCKSZ;
		baseValue = getOpt("base", baseValue);
		if (baseValue < CubesInterface.MIN_BASE ||
				baseValue > CubesInterface.MAX_BASE)
			baseValue = CubesInterface.DEFAULT_BASE;
		delayValue = getOpt("delay", delayValue);
	}

	public void showBuffer() {
		if (ioDialog != null) {
			ioDialog.setVisible(true);
		}
	}

	public void showMessage(String msg) {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Cubes", msg, "OK",
				mediumIcon, false);
			aDialog.setBounds(this.getSize().width / 2,
				this.getSize().height / 2, 450, 120);
			aDialog.setVisible(true);
		}
	}

	public void descriptionHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Description",
"Cubes - A 3D sliding block puzzle.  The cubes represented are" + NL +
"flattened out on the plane.  If the Z slider is greater than 1 then" + NL +
"each group of boxes is a layer of the puzzle.  If the Z slider = 1" + NL +
"then this is a simple 15 tile puzzle.  The 15 puzzle was originally" + NL +
"made around 1880 by Noyes Chapman (the 14-15 puzzle of Sam Loyd was" + NL +
"made soon after).  Similar 3D puzzles include BloxBox; Qrazy  Qube" + NL +
"(2x2x2) by P. Hein of Denmark and Varikon Box 'L (3x3x3)" + NL +
"(Hungary) and Mad Marbles (3x3x3) by Toys & Games Interna-" + NL +
"tional USA." + NL,
				"OK", largeIcon, true);
			aDialog.setBackground(Color.white);
			aDialog.setVisible(true);
		}
	}

	public void featuresHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Features",
"The puzzle works best with Num Lock on." + NL +
" " + NL +
"Click \"mouse-left\" button to move a block.  Clicks on spaces" + NL +
"or clicks on blocks that are not in line with a space will not" + NL +
"move the blocks.  Notice one can move more than one block at" + NL +
"a time by clicking on the block farthest from the space but" + NL +
"in line with the space.  Since you only have a 2-D screen, the" + NL +
"next level down is either the box to the right of the" + NL +
"current box or the bottom of the current box." + NL +
" " + NL +
"Double click \"mouse-right\" button, or press \"C\" or \"c\" keys," + NL +
"clear the puzzle." + NL +
" " + NL +
"Press \"g\" key to get (read) a saved puzzle to screen." + NL +
" " + NL +
"Press \"w\" key to save (write) a puzzle to screen." + NL +
" " + NL +
"Press \"G\" key to get (read) a saved puzzle to file." + NL +
" " + NL +
"Press \"W\" key to save (write) a puzzle to file." + NL +
" " + NL +
"Press \"U\" or \"u\" keys to undo a move." + NL +
" " + NL +
"Press \"R\" or \"r\" keys to redo a move." + NL +
" " + NL +
"Press \"Z\" or \"z\" keys to randomize.  New records can only be set" + NL +
"from a cleared state." + NL +
" " + NL +
"\"S\" or \"s\" keys start the auto-solver.  Only works where z = 1." + NL +
" " + NL +
"Press \"N\" or \"n\" keys to toggle the numbers or non-graphics mode." + NL +
"(Default is on)." + NL +
" " + NL +
"Press \"I\" or \"i\" keys to move block in (\"*\", \"5\" also work)." + NL +
" " + NL +
"Press \"O\" or \"o\" keys to move block out (\"/\", \"0\" also work)." + NL +
" " + NL +
"Press \">\" or \".\" keys to speed up the movement of the blocks." + NL +
" " + NL +
"Press \"<\" or \",\" keys to slow up the movement of the blocks." + NL +
" " + NL +
"Press \"@\" key to toggle the sound." + NL +
" " + NL +
"Press \"Esc\" key to hide program." + NL +
" " + NL +
"Press \"Q\", \"q\", or \"CTRL-C\" keys to kill program." + NL +
" " + NL +
"Use the key pad or arrow keys to move without the mouse." + NL +
"Key pad (better with Num Lock on) is defined for Cubes as:" + NL +
"    /    Out" + NL +
"    8    Up" + NL +
"    ^" + NL +
"4<5>6  Left, In, Right" + NL +
"    v" + NL +
"    2    Down" + NL +
"The arrow keys seem at first to work in the opposite direction" + NL +
"then you expect because the space is not the object moving." + NL +
" " + NL +
"The control key allows you to change the complexity of the" + NL +
"puzzle. Hold down control key and use the keypad or arrow" + NL +
"keys to reduce or enlarge puzzle complexity." + NL +
"Currently, there is no way to randomize the puzzle." + NL,
				"OK", largeIcon, true);
			aDialog.setBackground(Color.white);
			aDialog.setVisible(true);
		}
	}

	public void referencesHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"References",
"L. E. Horden, Sliding Piece Puzzles (Recreations in " +
"Mathematics Series), Oxford University Press 1986," + NL +
"pp 1, 157-159." + NL +
"Jerry Slocum & Jack Botermans, Puzzles Old & New (How to Make " +
"and Solve Them), University of Washington" + NL +
"Press, Seattle 1987, p 126, 127." + NL,
				"OK", mediumIcon, false);
			aDialog.setVisible(true);
		}
	}

	public void aboutHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"About",
"Java CubesFrame Version 8.4.0" + NL +
"Send bugs (reports or fixes) to the author:" +
"David Bagley <bagleyd AT verizon.net>" + NL +
"The latest version is at: " +
"https://www.sillycycle.com/puzzles.html" + NL +
"Mandrill and auto-solve from X puzzle by Don Bennett, HP Labs" + NL,
				"OK", mediumIcon, false);
			aDialog.setVisible(true);
		}
	}

	public Icon getIcon(String string, int sx, int sy) {
		URL url = CubesFrame.class.getResource("/" + string);

		if (url == null)
			return null;
		return new Icon(Toolkit.getDefaultToolkit().
			getImage(url), sx, sy);
	}

	// TODO Make more reliable, also should cache
	public void playClip(String string) {
		URL url = CubesFrame.class.getResource("/" + string);

		if (url == null)
			return;
		try {
			AudioInputStream audioInputStream =
				AudioSystem.getAudioInputStream(url);
			DataLine.Info info = new DataLine.Info(Clip.class,
				audioInputStream.getFormat());
			Clip clip = (Clip)AudioSystem.getLine(info);
			clip.open(audioInputStream);
			clip.start();
		} catch (IOException | LineUnavailableException | UnsupportedAudioFileException e) {
			e.printStackTrace();
		}
	}

	public void playBumpAudio() {
		playClip(CubesInterface.SOUNDS[CubesInterface.Sounds.BUMP.ordinal()]);

	}

	public void playDripAudio() {
		playClip(CubesInterface.SOUNDS[CubesInterface.Sounds.DRIP.ordinal()]);
	}

	@Override
	public void setSize(int width, int height) {
		if (puzzle == null)
			return;
		puzzle.setBounds(0, 0, width, height);
		/*repaint();*/ // causes flashing
	}

	@Override
	public void paint(Graphics g) {
		puzzle.setSize(secondaryPanel.getSize().width,
			secondaryPanel.getSize().height - topPanel.getSize().height);
		super.paint(g);
	}

	public String readUserScore() {
		StringBuilder scoreString = new StringBuilder();

		for (int i = 0; i < puzzleRecord.length; i++) {
			for (int j = i; j < puzzleRecord[i].length; j++) {
				for (int k = j; k < puzzleRecord[i][j].length; k++) {
					scoreString.append(puzzleRecord[i][j][k]);
					if (k < puzzleRecord[i][j].length - 1)

						scoreString.append(" ");
				}
				if (j < puzzleRecord[i].length - 1
						|| i < puzzleRecord.length - 1)
					scoreString.append(NL);
			}
			if (i < puzzleRecord.length - 1)
				scoreString.append(NL);
		}
		return scoreString.toString();
	}

	public void writeScore() {
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				SCOREFILE, false))) {
			pr.println(readUserScore());
		} catch (Exception e) {
			//e.printStackTrace();
		}
	}

	public void initRecords(String cookieValue, boolean store) {
		String[] st = null;
		int field = 0;

		if (cookieValue != null)
			st = cookieValue.split("[+/\\s]+");
		if (debug)
			System.out.println("initRecords: " + cookieValue);
		puzzleRecord = new int[MAX_BLOCKS - CubesInterface.MIN_BLOCKS + 1]
			[MAX_BLOCKS - CubesInterface.MIN_BLOCKS + 1]
			[MAX_BLOCKS - CubesInterface.MIN_BLOCKS + 1];
		for (int i = 0; i < puzzleRecord.length; i++) {
		  for (int j = i; j < puzzleRecord[i].length; j++) {
		    for (int k = j; k < puzzleRecord[i][j].length; k++) {
			if (st == null || st.length <= field) {
				//build it from the array
				puzzleRecord[k][j][i] =
					puzzleRecord[k][i][j] =
					puzzleRecord[j][k][i] =
					puzzleRecord[j][i][k] =
					puzzleRecord[i][k][j] =
					puzzleRecord[i][j][k] =
					CubesScores.PUZZLE_SCORES[field++];
			} else {
				puzzleRecord[k][j][i] =
					puzzleRecord[k][i][j] =
					puzzleRecord[j][k][i] =
					puzzleRecord[j][i][k] =
					puzzleRecord[i][k][j] =
					puzzleRecord[i][j][k] =
					Integer.parseInt(st[field++]);
			}
		    }
		  }
		}
		if (store && st == null)
			writeScore(); // if no cookie, then make one...
	}

	public void readScoreRecords() {
		try (BufferedReader br = new BufferedReader(
				new FileReader(SCOREFILE))) {
			StringBuilder scoreData = new StringBuilder();
			String lineData;

			while ((lineData = br.readLine()) != null) {
				scoreData.append(lineData);
				scoreData.append(NL);
			}
			initRecords(scoreData.toString(), true);
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Cannot open " + SCOREFILE);
			initRecords(null, true);
		}
	}

	public String readTextArea() {
		return ioDialog.getText();
	}

	public void writeTextArea(String s) {
		ioDialog.setText(s);
	}

	public void shuffleDown() {
		ComponentUtil.findJFrame(this).toBack();
	}

	public void quit() {
		System.exit(0);
	}

	void printRecord(int sizeX, int sizeY, int sizeZ) {
		int i = sizeX - CubesInterface.MIN_BLOCKS;
		int j = sizeY - CubesInterface.MIN_BLOCKS;
		int k = sizeZ - CubesInterface.MIN_BLOCKS;

		if (puzzleRecord == null)
			return;
		if (sizeX > MAX_BLOCKS || sizeY > MAX_BLOCKS ||
				sizeZ > MAX_BLOCKS) {
			recordValueLabel.setText("NONE");
		} else if (puzzleRecord[i][j][k] <= NEVER) {
			recordValueLabel.setText("NEVER");
		} else {
			recordValueLabel.setText(Integer.toString(
				puzzleRecord[i][j][k]));
		}
	}

	boolean handleSolved(int counter,
			int sizeX, int sizeY, int sizeZ) {
		int i = sizeX - CubesInterface.MIN_BLOCKS;
		int j = sizeY - CubesInterface.MIN_BLOCKS;
		int k = sizeZ - CubesInterface.MIN_BLOCKS;

		if (puzzleRecord != null &&
				sizeX <= MAX_BLOCKS &&
				sizeY <= MAX_BLOCKS &&
				sizeZ <= MAX_BLOCKS &&
				(counter < puzzleRecord[i][j][k] ||
				puzzleRecord[i][j][k] <= NEVER)) {
			/* Maybe its been updated by another */
			readScoreRecords();
			puzzleRecord[i][j][k] =
				puzzleRecord[i][k][j] =
				puzzleRecord[j][i][k] =
				puzzleRecord[j][k][i] =
				puzzleRecord[k][i][j] =
				puzzleRecord[k][j][i] =
				counter;
			writeScore();
			printRecord(sizeX, sizeY, sizeZ);
			return true;
		}
		return false;
	}

	public void initScore() {
		if (puzzleRecord == null) {
			readScoreRecords();
			printRecord(blocksValue.x, blocksValue.y, blocksValue.z);
		}
	}

	class BlockXChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = blocksValue.x;

			blocksValue.x = blockXChanger.getValue();
			if (oldValue != blocksValue.x) {
				puzzle.setPuzzleSizeX(blocksValue.x);
				//CubesCanvas.reformatPuzzle(blocksValue, blocksValue.x - oldValue);
				blockYChanger.setValue(blocksValue.y);
				blockZChanger.setValue(blocksValue.z);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				printRecord(blocksValue.x, blocksValue.y,
					blocksValue.z);
			}
		}
	}

	class BlockYChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = blocksValue.y;

			blocksValue.y = blockYChanger.getValue();
			if (oldValue != blocksValue.y) {
				puzzle.setPuzzleSizeY(blocksValue.y);
				//CubesCanvas.reformatPuzzle(blocksValue, blocksValue.y - oldValue);
				blockXChanger.setValue(blocksValue.x);
				blockZChanger.setValue(blocksValue.z);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				printRecord(blocksValue.x, blocksValue.y,
					blocksValue.z);
			}
		}
	}

	class BlockZChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = blocksValue.z;

			blocksValue.z = blockZChanger.getValue();
			if (oldValue != blocksValue.z) {
				puzzle.setPuzzleSizeZ(blocksValue.z);
				//CubesCanvas.reformatPuzzle(blocksValue, blocksValue.z - oldValue);
				blockXChanger.setValue(blocksValue.x);
				blockYChanger.setValue(blocksValue.y);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				printRecord(blocksValue.x, blocksValue.y,
					blocksValue.z);
			}
		}
	}

	class SpeedChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = speedValue;

			speedValue = speedChanger.getValue();
			if (oldValue != speedValue) {
				puzzle.setDelay(MAX_SPEED + MIN_SPEED - 1 -
					speedValue);
			}
		}
	}

	@Override
	public void actionPerformed(ActionEvent event) {
		String command = event.getActionCommand();

		for (int opt = 0; opt < COMMAND_STRINGS.length; opt++) {
			if (command.equals(COMMAND_STRINGS[opt])) {
				processKey(COMMAND_CHARS[opt],
					((event.getModifiers() &
					InputEvent.CTRL_DOWN_MASK) != 0));
				return;
			}
		}
	}

	public void callback(int reason) {
		messageDisplay = "                                                                                ";
		switch(reason) {
		case CubesInterface.ACTION_RESTORE:
		case CubesInterface.ACTION_RESET:
			movesDisplay = 0;
			randomized = false;
			break;
		case CubesInterface.ACTION_BLOCKED:
			messageDisplay = "Blocked";
			break;
		case CubesInterface.ACTION_SPACE:
			// Too annoying
			// messageDisplay = "A space cannot slide";
			break;
		case CubesInterface.ACTION_MOVED:
			movesDisplay++;
			break;
		case CubesInterface.ACTION_SOLVED:
			if (puzzle.getCheat())
				messageDisplay = "No cheating!!"; //username
			else if (!randomized && handleSolved(movesDisplay,
					blocksValue.x, blocksValue.y, blocksValue.z))
				messageDisplay = "Best solution, congratulations!!"; //username
			else
				messageDisplay = "Solved!";
			//showMessage(messageDisplay);
			break;
		case CubesInterface.ACTION_COMPUTED:
			break;
		case CubesInterface.ACTION_UNDO:
			movesDisplay--;
			break;
		case CubesInterface.ACTION_REDO:
			movesDisplay++;
			break;
		case CubesInterface.ACTION_RANDOMIZE:
			movesDisplay = 0;
			randomized = true;
			break;
		case CubesInterface.ACTION_SPEED:
			{
				int oldDelay, delay;

				delay = puzzle.getDelay();
				oldDelay = delay;
				if (delay > MAX_SPEED - MIN_SPEED)
					delay = MAX_SPEED - MIN_SPEED;
				if (delay < 0)
					delay = 0;
				if (delay != oldDelay) {
					puzzle.setDelay(delay);
				}
				if (speedChanger != null)
					speedChanger.setValue(MAX_SPEED +
						MIN_SPEED - 1 - delay);
			}
			return;
		case CubesInterface.ACTION_INCX:
			movesDisplay = 0;
			blocksValue.x++;
			printRecord(blocksValue.x, blocksValue.y, blocksValue.z);
			puzzle.setPuzzleSizeX(blocksValue.x);
			//CubesCanvas.reformatPuzzle(blocksValue, 1);
			if (blockXChanger != null)
				blockXChanger.setValue(blocksValue.x);
			if (blockYChanger != null)
				blockYChanger.setValue(blocksValue.y);
			if (blockZChanger != null)
				blockZChanger.setValue(blocksValue.z);
			break;
		case CubesInterface.ACTION_DECX:
			movesDisplay = 0;
			blocksValue.x--;
			printRecord(blocksValue.x, blocksValue.y, blocksValue.z);
			puzzle.setPuzzleSizeX(blocksValue.x);
			//CubesCanvas.reformatPuzzle(blocksValue, -1);
			if (blockXChanger != null)
				blockXChanger.setValue(blocksValue.x);
			if (blockYChanger != null)
				blockYChanger.setValue(blocksValue.y);
			if (blockZChanger != null)
				blockZChanger.setValue(blocksValue.z);
			break;
		case CubesInterface.ACTION_INCY:
			movesDisplay = 0;
			blocksValue.y++;
			printRecord(blocksValue.x, blocksValue.y, blocksValue.z);
			puzzle.setPuzzleSizeY(blocksValue.y);
			//CubesCanvas.reformatPuzzle(blocksValue, 1);
			if (blockXChanger != null)
				blockXChanger.setValue(blocksValue.x);
			if (blockYChanger != null)
				blockYChanger.setValue(blocksValue.y);
			if (blockZChanger != null)
				blockZChanger.setValue(blocksValue.z);
			break;
		case CubesInterface.ACTION_DECY:
			movesDisplay = 0;
			blocksValue.y--;
			printRecord(blocksValue.x, blocksValue.y, blocksValue.z);
			puzzle.setPuzzleSizeY(blocksValue.y);
			//CubesCanvas.reformatPuzzle(blocksValue, -1);
			if (blockXChanger != null)
				blockXChanger.setValue(blocksValue.x);
			if (blockYChanger != null)
				blockYChanger.setValue(blocksValue.y);
			if (blockZChanger != null)
				blockZChanger.setValue(blocksValue.z);
			break;
		case CubesInterface.ACTION_INCZ:
			movesDisplay = 0;
			blocksValue.z++;
			printRecord(blocksValue.x, blocksValue.y, blocksValue.z);
			puzzle.setPuzzleSizeZ(blocksValue.z);
			//CubesCanvas.reformatPuzzle(blocksValue, 1);
			if (blockXChanger != null)
				blockXChanger.setValue(blocksValue.x);
			if (blockYChanger != null)
				blockYChanger.setValue(blocksValue.y);
			if (blockZChanger != null)
				blockZChanger.setValue(blocksValue.z);
			break;
		case CubesInterface.ACTION_DECZ:
			movesDisplay = 0;
			blocksValue.z--;
			printRecord(blocksValue.x, blocksValue.y, blocksValue.z);
			puzzle.setPuzzleSizeZ(blocksValue.z);
			//CubesCanvas.reformatPuzzle(blocksValue, -1);
			if (blockXChanger != null)
				blockXChanger.setValue(blocksValue.x);
			if (blockYChanger != null)
				blockYChanger.setValue(blocksValue.y);
			if (blockZChanger != null)
				blockZChanger.setValue(blocksValue.z);
			break;
		case CubesInterface.ACTION_IGNORE:
			messageDisplay = "Clear to start";
			break;
		default:
		}
		messageLabel.setText(messageDisplay);
		movesValueLabel.setText(Integer.toString(movesDisplay));
	}

	public CubesFrame(String title) {
		/*JPanel info1Panel = new JPanel();
		JPanel leftPanel = new JPanel();
		JPanel rightPanel = new JPanel();*/
		JPanel startPanel = new JPanel();
		JPanel info2Panel = new JPanel();
		JPanel movesPanel = new JPanel();
		JPanel recordPanel = new JPanel();
		JPanel sliderPanel = new JPanel();
		JPanel canvasPanel = new JPanel();

		blocksValue.x = CubesInterface.DEFAULT_BLOCKSX;
		blocksValue.y = CubesInterface.DEFAULT_BLOCKSY;
		blocksValue.z = CubesInterface.DEFAULT_BLOCKSZ;
		try {
			mediumIcon = getIcon(CubesInterface.ICONS[CubesInterface.Icons.X32.ordinal()], 32, 32);
			largeIcon = getIcon(CubesInterface.ICONS[CubesInterface.Icons.X48.ordinal()], 48, 48);
		} catch (Exception e) {
			largeIcon = null;
		}
		ioDialog = new IODialog(ComponentUtil.findJFrame(this),
			CubesInterface.TITLE + ": IO", "Puzzle Position",
			"DONE");
		readScoreRecords();
		getOpts();
		speedValue = MAX_SPEED + MIN_SPEED - 1 - delayValue;
		mainPanel.setLayout(new BorderLayout());
		mainPanel.setBackground(Color.white);
		if (controlValue) {
			menuBar = new JMenuBar();
			JMenu menu;
			JMenuItem menuItem;

			JPopupMenu.setDefaultLightWeightPopupEnabled(false);
			menu = new JMenu("File");
			menuItem = new JMenuItem("Get Buffer");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_G);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Write Buffer");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_W);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Show Buffer");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menu.addSeparator();
			menuItem = new JMenuItem("Get File");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Write File");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menu.addSeparator();
			menuItem = new JMenuItem("Exit");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_X);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			menu = new JMenu("Play");
			menuItem = new JMenuItem("Undo");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_U);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Redo");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_R);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Clear");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_C);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Randomize");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_Z);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Auto-solve");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_S);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Number");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_N);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Speed >");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Slow <");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			soundMenuItem = new JCheckBoxMenuItem("Sound @", sound);
			menu.add(soundMenuItem);
			soundMenuItem.addActionListener(this);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			menu = new JMenu("Help");
			menuItem = new JMenuItem("Description ?");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Features !");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("References ^");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("About");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_A);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			//setJMenuBar(menuBar); // needs to be in a panel
			mainPanel.add(menuBar, BorderLayout.PAGE_START);
			/*leftPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
			leftPanel.add(new JLabel("Move block"));
			rightPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
			rightPanel.add(new JLabel("Clear"));
			leftPanel.add(getIcon("mouse-l.png", 16, 16));
			rightPanel.add(getIcon("mouse-r.png", 16, 16));
			info1Panel.setLayout(new FlowLayout(FlowLayout.LEFT, 15, 0));
			info1Panel.add(leftPanel);
			info1Panel.add(rightPanel);*/
			movesPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 0));
			movesPanel.add(new JLabel("Moves"));
			movesValueLabel.setColumns(5);
			movesValueLabel.setEditable(false);
			movesPanel.add(movesValueLabel);
			recordPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 0));
			recordPanel.add(new JLabel("Record"));
			recordValueLabel.setColumns(5);
			recordValueLabel.setEditable(false);
			recordPanel.add(recordValueLabel);
			info2Panel.setLayout(new FlowLayout(FlowLayout.LEFT, 15, 5));
			info2Panel.add(movesPanel);
			info2Panel.add(recordPanel);
			sliderPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 5));
			//CubesCanvas.reformatPuzzle(blocksValue, 0);
			blockXChanger = new Spinner("Blocks X",
				blocksValue.x, CubesInterface.MIN_BLOCKS);
			blockYChanger = new Spinner("Blocks Y",
				blocksValue.y, CubesInterface.MIN_BLOCKS);
			blockZChanger = new Spinner("Blocks Z",
				blocksValue.z, CubesInterface.MIN_BLOCKS);
			/*blockXChanger = new Slider("Blocks X",
				blocksValue.x, CubesInterface.MIN_BLOCKS,
				((blocksValue.x > MAX_BLOCKS) ? blocksValue.x : MAX_BLOCKS));
			blockYChanger = new Slider("Blocks Y",
				blocksValue.y, CubesInterface.MIN_BLOCKS,
				((blocksValue.y > MAX_BLOCKS) ? blocksValue.y : MAX_BLOCKS));
			blockZChanger = new Slider("Blocks Z",
				blocksValue.z, CubesInterface.MIN_BLOCKS,
				((blocksValue.z > MAX_BLOCKS) ? blocksValue.z : MAX_BLOCKS));*/
			sliderPanel.add(blockXChanger);
			sliderPanel.add(blockYChanger);
			sliderPanel.add(blockZChanger);
			topPanel.setLayout(new ColumnLayout(0, 0, 0,
				ColumnLayout.CENTER));
			/*topPanel.add(info1Panel);*/
			topPanel.add(info2Panel);
			topPanel.add(sliderPanel);
			speedChanger = new Spinner("Animation Speed",
				speedValue, MIN_SPEED,
				((speedValue > MAX_SPEED) ? speedValue : MAX_SPEED));
			/*speedChanger = new Slider("Animation Speed",
				speedValue, MIN_SPEED,
				((speedValue > MAX_SPEED) ? speedValue : MAX_SPEED));*/
			topPanel.add(speedChanger);
			//speedChanger.setWidth(150);
			topPanel.add(messageLabel);
			startPanel.setLayout(new ColumnLayout(0, 0, 0,
				ColumnLayout.LEFT));
			startPanel.add(topPanel);
			secondaryPanel.setLayout(new BorderLayout());
			secondaryPanel.add(startPanel, BorderLayout.PAGE_START);
		}
		puzzle = new Cubes(this,
			blocksValue.x, blocksValue.y, blocksValue.z,
			baseValue,
			blockColor, fgColor, bgColor,
			MAX_SPEED + MIN_SPEED - 1 - speedValue);
		canvasPanel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.LEFT));
		canvasPanel.add(puzzle);
		secondaryPanel.add(canvasPanel, BorderLayout.CENTER);
		mainPanel.add(secondaryPanel, BorderLayout.CENTER);
		getContentPane().add(mainPanel, BorderLayout.CENTER);
		if (controlValue) {
			printRecord(blocksValue.x, blocksValue.y, blocksValue.z);
			blockXChanger.addChangeListener(new BlockXChangeListener());
			blockYChanger.addChangeListener(new BlockYChangeListener());
			blockZChanger.addChangeListener(new BlockZChangeListener());
			speedChanger.addChangeListener(new SpeedChangeListener());
		}
		canvasPanel.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseEntered(MouseEvent event) {
				puzzle.requestFocus();
			}
		});
		setTitle(title);
		/* puzzle.setSize(getSize().width,
			getSize().height - topPanel.getSize().height); */
		puzzle.setSize(widthValue, heightValue);
	}

	public static void main(String[] args) {
		String laf = UIManager.getCrossPlatformLookAndFeelClassName();

		try {
			UIManager.setLookAndFeel(laf);
		} catch (ClassNotFoundException | InstantiationException | IllegalAccessException | UnsupportedLookAndFeelException e) {
			e.printStackTrace();
			System.exit(1);
		}
		List<Image> iconList = new ArrayList<>();
		for (String icon : CubesInterface.ICONS) {
			URL url = CubesFrame.class.getResource("/" + icon);
			if (url != null)
				iconList.add(Toolkit.getDefaultToolkit().getImage(url));
		}

		argumentParser = new ArgumentParser(args);
		if (argumentParser.hasOption("?")) {
			System.out.println("usage:" + NL +
"\t[-windowWidth={int}] [-windowHeight={int}]" + NL +
"\t[-windowX={int}] [-windowY={int}]" + NL +
"\t[-blockColor={int}] [-fgColor={int}] [-bgColor={int}]" + NL +
"\t[-sound={bool}] [-control={bool}]" + NL +
"\t[-size{x|y|z}={int}] [-base={int}] [-delay={int}]");
			return;
		}
		CubesFrame frame = new CubesFrame(CubesInterface.TITLE);
		frame.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent e) {
				System.exit(0);
			}
		});
		if (!iconList.isEmpty()) {
			frame.setIconImages(iconList);
		}
		frame.pack();
		frame.puzzle.requestFocusInWindow();
		frame.setLocation(CubesFrame.xValue, CubesFrame.yValue);
		frame.setVisible(true);
	}
}
